# -*- coding: utf-8 -*-
"""Static playground.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import ABC, abstractmethod
from typing import List

import numpy as np

from auxein.fitness import Fitness
from auxein.population import Population, Individual


class Replacement(ABC):

    def __init__(self, offspring_size: int) -> None:
        self.offspring_size = offspring_size

    def _replace(self, quantity: int, offspring: List[Individual], population: Population, individuals_to_kill: List[str], fitness_function: Fitness) -> None:
        for i in individuals_to_kill:
            population.kill(i)

        children: List[Individual] = np.random.choice(offspring, quantity, replace=False)
        for child in children:
            population.add(child, fitness_function.fitness(child))

    @abstractmethod
    def replace(self, offspring: List[Individual], population: Population, fitness_function: Fitness) -> None:
        pass


class ReplaceWorst(Replacement):

    def __init__(self, offspring_size: int) -> None:
        super().__init__(offspring_size=offspring_size)

    def replace(self, offspring: List[Individual], population: Population, fitness_function: Fitness) -> None:
        quantity = population.size() if self.offspring_size >= population.size() else min(self.offspring_size, len(offspring))
        individuals_to_kill: List[str] = list(map(lambda item : item[0], population.rank_by_fitness(quantity, reverse=False)))
        super()._replace(quantity, offspring, population, individuals_to_kill, fitness_function)



"""
Contains various genotypes representations.
"""
from __future__ import absolute_import

import numpy as np


class Genotype:

    def __init__(self, dna: np.ndarray, mask: np.ndarray) -> None:
        self._dimension = len(dna)
        self._dna = dna.copy()
        self._mask = mask.copy()

    @property
    def dimension(self) -> int:
        return self._dimension

    @property
    def dna(self) -> np.ndarray:
        return self._dna.copy()

    @property
    def mask(self) -> np.ndarray:
        return self._mask.copy()

    def __repr__(self) -> str:
        repr: str = f'({self._dna}),({self._mask})'
        return repr



# flake8: noqa
from .genotype import Genotype

from .dna_builders import UniformRandomDnaBuilder, NormalRandomDnaBuilder

from .individual import Individual
from .individual import build_individual

from .core import build_fixed_dimension_population, build_variable_dimension_population, Item, Population



"""Contains few classes to build random dna sequences.
"""
from __future__ import absolute_import
from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np


class DnaBuilder(ABC):

    @abstractmethod
    def get(self, dimension: int) -> np.ndarray:
        pass


class RandomDnaBuilder(DnaBuilder):

    def __init__(self, distribution: str):
        super().__init__()
        self._distribution = distribution

    def get_distribution(self) -> str:
        return self._distribution

    @abstractmethod
    def get(self, dimension: int) -> np.ndarray:
        pass


class UniformRandomDnaBuilder(RandomDnaBuilder):

    def __init__(self, interval: Tuple[float, float] = (-1.0, 1.0)):
        super().__init__(distribution='uniform')
        self.interval = interval

    def get(self, dimension: int) -> np.ndarray:
        assert dimension > 0, 'dna dimension must be strictly positive.'
        return np.random.uniform(
            self.interval[0],
            self.interval[1],
            dimension
        )


class NormalRandomDnaBuilder(RandomDnaBuilder):

    def __init__(self, mean: float = 0.0, std: float = 1.0):
        super().__init__(distribution='normal')
        self.mean = mean
        self.std = std

    def get(self, dimension: int) -> np.ndarray:
        assert dimension > 0, 'dna dimension must be strictly positive.'
        return np.random.normal(
            self.mean,
            self.std,
            dimension
        )



# -*- coding: utf-8 -*-
"""SimpleArithmetic.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import ABC, abstractmethod
from typing import Tuple, List

import numpy as np


class Recombination(ABC):

    def __init__(self, allow_uneven: bool = False):
        self.allow_uneven = allow_uneven

    @abstractmethod
    def recombine(self, parent1_dna: np.ndarray, parent2_dna: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        pass


class SimpleArithmetic(Recombination):

    def __init__(self, alpha: float, allow_uneven: bool = False) -> None:
        super().__init__(allow_uneven=allow_uneven)
        self.alpha = alpha

    def __linear_combination(self, arr1: np.ndarray, arr2: np.ndarray) -> List[float]:
        result: List[float] = []
        for (i1, i2) in zip(arr1, arr2):
            result.append(self.alpha * i2 + (1 - self.alpha) * i1)
        return result

    def recombine(self, parent1_dna: np.ndarray, parent2_dna: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        if not self.allow_uneven:
            assert len(parent1_dna) == len(parent2_dna), 'dna of parents must be of the same dimension'

        cross_over_point = np.random.randint(0, min(len(parent1_dna), len(parent2_dna)))
        child1_dna = np.concatenate(
            (
                parent1_dna[:cross_over_point],
                self.__linear_combination(
                    parent1_dna[cross_over_point:],
                    parent2_dna[cross_over_point:]
                )
            )
        )
        if self.allow_uneven and len(child1_dna) < len(parent1_dna):
            child1_dna = np.concatenate((child1_dna, parent1_dna[len(child1_dna):]))

        child2_dna = np.concatenate(
            (
                parent2_dna[:cross_over_point],
                self.__linear_combination(
                    parent2_dna[cross_over_point:],
                    parent1_dna[cross_over_point:]
                )
            )
        )
        if self.allow_uneven and len(child2_dna) < len(parent2_dna):
            child2_dna = np.concatenate((child2_dna, parent2_dna[len(child2_dna):]))

        return (child1_dna, child2_dna)


class MatrixRecombination(Recombination):

    def __init__(self, shape: Tuple[int, int], recombination: Recombination) -> None:
        self._shape = shape
        self.recombination = recombination

    def __vectorise(self, matrix: np.ndarray) -> np.ndarray:
        (r, c) = matrix.shape
        return matrix.reshape((1, r * c))

    def __to_matrix(self, vector: np.ndarray) -> np.ndarray:
        return vector.reshape(self._shape)

    def recombine(self, parent1_dna: np.ndarray, parent2_dna: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        vectorised_parent1_dna = self.__vectorise(parent1_dna)
        vectorised_parent2_dna = self.__vectorise(parent2_dna)
        (vectorised_child_1, vectorised_child_2) = self.recombination.recombine(
            vectorised_parent1_dna,
            vectorised_parent2_dna
        )
        return (
            self.__to_matrix(vectorised_child_1),
            self.__to_matrix(vectorised_child_2)
        )



"""Contains the base Individual class.
"""
from __future__ import absolute_import
from typing import List, Optional
from uuid import uuid4, UUID
import time

import numpy as np

from auxein.mutations import Mutation
from auxein.population.genotype import Genotype


class Individual:

    def __init__(self, genotype: Genotype, id: Optional[str] = None) -> None:
        self._id = uuid4() if id is None else UUID(id)
        self._born_at = time.time()
        self._genotype = genotype

    @property
    def id(self) -> str:
        return str(self._id)

    def age(self) -> float:
        return time.time() - self._born_at

    def dimension(self) -> int:
        return self._genotype.dimension

    @property
    def genotype(self) -> Genotype:
        return self._genotype

    def mutate(self, mutation_function: Mutation) -> 'Individual':
        return Individual(mutation_function.mutate(self._genotype))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Individual):
            return False
        return self._id == other._id

    def __hash__(self) -> int:
        return hash(self._id)

    def __repr__(self) -> str:
        return f'[{self._id}],({self._genotype})'


def build_individual(dna: List[float], mask: List[float] = [], id: Optional[str] = None) -> Individual:
    """Utility function to build an Individual."""
    return Individual(Genotype(np.array(dna), np.array(mask)), id)



# -*- coding: utf-8 -*-
"""Static playground.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import ABC, abstractmethod
from typing import Dict, Any, List, Tuple

import logging
from itertools import permutations

import numpy as np

from auxein.fitness import Fitness
from auxein.population.individual import build_individual, Individual
from auxein.population import Population
from auxein.mutations import Mutation
from auxein.recombinations import Recombination
from auxein.parents.distributions import Distribution
from auxein.parents.selections import Selection
from auxein.replacements import Replacement

logging.basicConfig(level=logging.DEBUG)


class Playground(ABC):

    def __init__(self, population: Population, fitness: Fitness) -> None:
        self.population = population
        self.fitness = fitness

    @abstractmethod
    def train(self, max_generations: int) -> Dict[str, Any]:
        pass

    @abstractmethod
    def predict(self, x: np.ndarray) -> float:
        pass

    def _get_nth_top_performant(self, depth: int = 0) -> Individual:
        id = self.population.rank_by_fitness(depth + 1)[depth][0]
        return self.population.get(id).individual


class Static(Playground):

    def __init__(
        self,
        population: Population,
        fitness: Fitness,
        mutation: Mutation,
        distribution: Distribution,
        selection: Selection,
        recombination: Recombination,
        replacement: Replacement,
        verbose: bool = False
    ) -> None:
        super().__init__(population=population, fitness=fitness)
        self.mutation = mutation
        self.distribution = distribution
        self.selection = selection
        self.recombination = recombination
        self.replacement = replacement
        self.verbose = verbose

    def __mate(self, mating_pool: List[str]) -> List[Individual]:
        couples = permutations(mating_pool, 2)
        offspring = []
        for (parent1_id, parent2_id) in couples:
            (child1_genotype_dna, child1_genotype_mask, child2_genotype_dna, child2_genotype_mask) = self.__breed(
                parent1_id,
                parent2_id
            )

            offspring.append(build_individual(child1_genotype_dna, child1_genotype_mask))
            offspring.append(build_individual(child2_genotype_dna, child2_genotype_mask))

        return offspring

    def __breed(self, parent1_id: str, parent2_id: str) -> Tuple[List[float], List[float], List[float], List[float]]:
        parent_1 = self.population.get(parent1_id).individual.mutate(self.mutation)
        parent_2 = self.population.get(parent2_id).individual.mutate(self.mutation)

        parent1_genotype_dna = parent_1.genotype.dna
        parent2_genotype_dna = parent_2.genotype.dna

        (child1_genotype_dna, child2_genotype_dna) = self.recombination.recombine(
            parent1_genotype_dna,
            parent2_genotype_dna
        )
        return (child1_genotype_dna, parent_1.genotype.mask, child2_genotype_dna, parent_2.genotype.mask)

    def train(self, max_generations: int) -> Dict[str, Any]:
        logging.info(f'Starting evolution cycle with a maximum of {max_generations} generations')
        stats: Dict[str, Any] = {
            'generations': {}
        }
        while self.population.generation_count < max_generations and self.population.size() > self.selection.offspring_size:
            mean_fitness = self.population.mean_fitness()
            if self.verbose is True:
                logging.debug(f'Running generation: {self.population.generation_count}/{max_generations} -- average_fitness: {mean_fitness} -- population size: {self.population.size()}')
            else:
                logging.debug(f'Running generation: {self.population.generation_count}/{max_generations}')

            stats['generations'][self.population.generation_count] = {}
            stats['generations'][self.population.generation_count]['mean_fitness'] = mean_fitness
            stats['generations'][self.population.generation_count]['genome'] = self.population.get_full_genome()

            # Mating step
            distribution = self.distribution.get(self.population)
            individual_ids = list(map(lambda i : i[0], distribution))
            probabilities = list(map(lambda i : i[1], distribution))

            mating_pool: List[str] = self.selection.select(individual_ids, probabilities)
            offspring = self.__mate(mating_pool)

            # Replacement step
            self.replacement.replace(offspring, self.population, self.fitness)
            self.population.update(self.fitness)

        logging.info(f'Training ended with average_fitness: {self.population.mean_fitness()} and a population size of {self.population.size()}')
        return stats

    def predict(self, x: np.ndarray, depth: int = 0) -> float:
        i = super()._get_nth_top_performant(depth)
        return self.fitness.value(i, x)

    def get_most_performant(self, depth: int = 0) -> Individual:
        return super()._get_nth_top_performant(depth)



# flake8: noqa

from .core import Fitness
from .kernel_based import GlobalMinimum
from .observation_based import ObservationBasedFitness, MultipleLinearRegression, SimplePolynomialRegression, MultipleLinearRegression


# flake8: noqa

from __future__ import absolute_import

from .population import Genotype
from .population import Individual
from .population import Item, Population

from .playgrounds import Static
from .mutations import Uniform
from .recombinations import SimpleArithmetic

from .parents import distributions
from .parents import selections

from .replacements import ReplaceWorst

from .fitness import MultipleLinearRegression


"""Contains the population class.
"""
from __future__ import absolute_import
from copy import deepcopy
from typing import Tuple, NamedTuple, Iterable, Dict, Any, Optional, List

import numpy as np

from auxein.population.dna_builders import DnaBuilder
from auxein.population.individual import Individual, build_individual
from auxein.fitness.core import Fitness

Item = NamedTuple('Item', [('individual', Individual), ('fitness', float)])


class Population:

    def __init__(self) -> None:
        self.__pool: Dict[str, Item] = {}
        self.__generation_count = 0

    def size(self) -> int:
        return len(self.__pool)

    def add(self, individual: Individual, fitness: float) -> None:
        self.__pool[str(individual.id)] = Item(deepcopy(individual), fitness)

    def get(self, individual_id: str) -> Item:
        return deepcopy(self.__pool[individual_id])

    def update(self, fitness_function: Fitness) -> None:
        for item in self.__pool.values():
            individual = item.individual
            new_fitness = fitness_function.fitness(individual)
            self.__pool[str(individual.id)] = Item(
                deepcopy(individual), new_fitness)
        self.__generation_count += 1

    def kill(self, individual_id: str) -> None:
        try:
            del self.__pool[individual_id]
        except KeyError:
            pass

    @property
    def pool(self) -> Iterable[Item]:
        return self.__pool.values()

    def total_fitness(self) -> float:
        total_fitness: float = 0.0
        for item in self.__pool.values():
            total_fitness += item.fitness
        return total_fitness

    @property
    def generation_count(self) -> int:
        return self.__generation_count

    def rank_by_fitness(self, k: Optional[int] = None, reverse: bool = True) -> List[Tuple[str, float]]:
        sorted_values = sorted(self.pool, key=lambda i: i.fitness, reverse=reverse)
        return list(map(lambda item : (item.individual.id, item.fitness), sorted_values))[:k]

    def __get_ages(self) -> List[float]:
        return list(map(lambda item : item.individual.age(), list(self.__pool.values())))

    def __get_fitness(self) -> List[float]:
        return list(map(lambda item : item.fitness, list(self.__pool.values())))

    def mean_age(self) -> float:
        pools_ages = self.__get_ages()
        value: float = np.mean(pools_ages)
        return value

    def std_age(self) -> float:
        pools_ages = self.__get_ages()
        value: float = np.std(pools_ages)
        return value

    def max_age(self) -> float:
        pools_ages = self.__get_ages()
        value: float = np.max(pools_ages)
        return value

    def min_age(self) -> float:
        pools_ages = self.__get_ages()
        value: float = np.min(pools_ages)
        return value

    def mean_fitness(self) -> float:
        return self.total_fitness() / self.size()

    def max_fitness(self) -> float:
        pools_fitness_values = self.__get_fitness()
        value: float = np.max(pools_fitness_values)
        return value

    def min_fitness(self) -> float:
        pools_fitness_values = self.__get_fitness()
        value: float = np.min(pools_fitness_values)
        return value

    def std_fitness(self) -> float:
        pools_fitness_values = self.__get_fitness()
        value: float = np.std(pools_fitness_values)
        return value

    def get_stats(self) -> Dict[str, Any]:
        return {
            'generation_count': self.__generation_count,
            'size': self.size(),
            'mean_age': self.mean_age(),
            'std_age': self.std_age(),
            'max_age': self.max_age(),
            'min_age': self.min_age(),
            'mean_fitness': self.mean_fitness(),
            'min_fitness': self.min_fitness(),
            'max_fitness': self.max_fitness(),
            'std_fitness': self.std_fitness()
        }

    def get_full_genome(self) -> np.ndarray:
        genome = []
        for item in self.pool:
            genome.append(item.individual.genotype.dna)
        return np.array(genome)


def build_random_individual(dimension: int, dna_builder: DnaBuilder) -> Individual:
    mask = np.repeat(np.random.normal(0, 1), dimension)
    dna = dna_builder.get(dimension)
    return build_individual(dna, mask)


def __add_to_population(population: Population, dimension: int, fitness_function: Fitness, dna_builder: DnaBuilder) -> None:
    individual = build_random_individual(dimension, dna_builder)
    fitness = fitness_function.fitness(individual)
    population.add(individual, fitness)


def build_fixed_dimension_population(dimension: int, initial_size: int, fitness_function: Fitness, dna_builder: DnaBuilder) -> Population:
    population = Population()
    for _ in range(0, initial_size):
        __add_to_population(population, dimension, fitness_function, dna_builder)
    return population


def build_variable_dimension_population(initial_size: int, fitness_function: Fitness, dna_builder: DnaBuilder) -> Population:
    population = Population()
    for _ in range(0, initial_size):
        dimension = np.random.randint(1, 10)
        __add_to_population(population, dimension, fitness_function, dna_builder)
    return population



# -*- coding: utf-8 -*-
"""Core Auxein mutations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Callable

import numpy as np

from .core import Fitness
from auxein.population import Individual


class GlobalMinimum(Fitness):

    def __init__(self, kernel: Callable[[np.ndarray], float]) -> None:
        super().__init__()
        self.kernel = kernel

    def fitness(self, individual: Individual) -> float:
        dna = individual.genotype.dna
        return -1 * self.kernel(dna)

    def value(self, individual: Individual, x: np.ndarray) -> float:
        return self.kernel(x)



# -*- coding: utf-8 -*-
"""Core Auxein mutations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import ABC, abstractmethod

from auxein.population.genotype import Genotype

import numpy as np


class Mutation(ABC):

    def __init__(self, extend_probability: float = 0.0):
        assert 0 <= extend_probability <= 1, 'extend_probability must be within [0, 1]'
        self.extend_probability = extend_probability

    def _extend(self, genotype: Genotype, new_gene: float) -> Genotype:
        if np.random.uniform(0, 1) <= self.extend_probability:
            dna = genotype.dna
            mask = genotype.mask
            return Genotype(np.append(dna, new_gene), np.append(mask, np.random.normal(0, 1)))
        return genotype

    @abstractmethod
    def mutate(self, genotype: Genotype) -> Genotype:
        pass


class Uniform(Mutation):

    def __init__(self, lower_bound: float, upper_bound: float, extend_probability: float = 0.0) -> None:
        super().__init__(extend_probability=extend_probability)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def mutate(self, genotype: Genotype) -> Genotype:
        gene_index = np.random.randint(
            0,
            genotype.dimension
        )
        dna = genotype.dna
        dna[gene_index] = np.random.uniform(self.lower_bound, self.upper_bound)
        new_gene = np.random.uniform(self.lower_bound, self.upper_bound)
        return super()._extend(Genotype(dna, genotype.mask), new_gene)


class FixedVariance(Mutation):

    def __init__(self, sigma: float, extend_probability: float = 0.0) -> None:
        super().__init__(extend_probability=extend_probability)
        self.sigma = sigma

    def mutate(self, genotype: Genotype) -> Genotype:
        move = np.vectorize(lambda g: g + np.random.normal(0, self.sigma))
        new_gene = np.random.normal(0, self.sigma)
        return super()._extend(Genotype(move(genotype.dna), genotype.mask), new_gene)


class SelfAdaptiveSingleStep(Mutation):

    def __init__(self, tau: float, extend_probability: float = 0.0) -> None:
        super().__init__(extend_probability=extend_probability)
        self.tau = tau

    def mutate(self, genotype: Genotype) -> Genotype:
        multiplier = np.exp(np.random.normal(0, self.tau))
        move_mask = np.vectorize(lambda g: g * multiplier)
        updated_mask = move_mask(genotype.mask)
        scalr = np.random.normal(0, 1, genotype.dimension)
        dna = genotype.dna + (updated_mask * scalr)
        new_gene = np.random.normal(0, self.tau)
        return super()._extend(Genotype(dna, updated_mask), new_gene)



# -*- coding: utf-8 -*-
"""Core Auxein mutations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Callable

import numpy as np


def linear_fit(coeff: np.ndarray, x: np.ndarray) -> float:
    assert type(coeff) == type(x) == np.ndarray, 'coefficients and variable must be np.ndarray'
    assert coeff.size - 1 == x.size, 'coefficients must be of the size of x+1'
    value: float = np.dot(x, coeff[:coeff.size - 1]) + coeff[coeff.size - 1]
    return value


def polynomial_fit(coeff: np.ndarray, x: np.ndarray) -> float:
    assert type(coeff) and x.size == 1, 'Only simple polynomial fit is supported.'
    result: float = np.polyval(coeff, x[0])
    return result


def residual(coeff: np.array, x: np.array, yi: float, fit: Callable[[np.ndarray, np.ndarray], float] = linear_fit) -> float:
    return (yi - fit(coeff, x))**2


def least_squares(xs: np.ndarray, y: np.ndarray, coeff: np.ndarray, fit: Callable[[np.ndarray, np.ndarray], float] = linear_fit) -> float:
    lsm: float = 0
    for x, yi in zip(xs, y):
        lsm = lsm + residual(coeff, x, yi, fit)
    return lsm


def logit(alpha: float, coeff: np.ndarray, x: np.ndarray) -> float:
    kernel: float = 0
    for (bi, xi) in zip(coeff, x):
        kernel += bi * xi
    result: float = 1 / (1 + np.exp(-(alpha + kernel)))
    return result



# -*- coding: utf-8 -*-
"""Core Auxein mutations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import abstractmethod

import numpy as np

from .core import Fitness
from .utils import linear_fit, polynomial_fit, least_squares, logit
from auxein.population import build_individual, Individual


class ObservationBasedFitness(Fitness):

    @abstractmethod
    def fitness(self, individual: Individual) -> float:
        pass

    @abstractmethod
    def value(self, individual: Individual, x: np.ndarray) -> float:
        pass

    def __compute_f(self, A: np.ndarray, B: np.ndarray) -> np.ndarray:
        F = []
        for v in np.transpose([np.tile(A, len(B)), np.repeat(B, len(A))]):
            F.append([v[0], v[1], self.fitness(build_individual([v[0], v[1]]))])
        return np.array(F)

    def get_landscape(self, specs: np.ndarray, size: int) -> np.ndarray:
        assert len(specs) == 2, 'Only 2-dimensional fitness landscapes are supported at the moment.'
        A = np.linspace(specs[0][0], specs[0][1], size)
        B = np.linspace(specs[1][0], specs[1][1], size)
        return self.__compute_f(A, B)


class MultipleLinearRegression(ObservationBasedFitness):

    def __init__(self, xs: np.ndarray, y: np.ndarray) -> None:
        super().__init__()
        assert xs.shape == (y.shape[0], xs.shape[1]), 'length of xs must be equal to length of y'
        self.xs = xs
        self.y = y

    def fitness(self, individual: Individual) -> float:
        dna = individual.genotype.dna
        return -1 * least_squares(self.xs, self.y, dna)

    def value(self, individual: Individual, x: np.ndarray) -> float:
        dna = individual.genotype.dna
        return linear_fit(dna, x)


class SimplePolynomialRegression(ObservationBasedFitness):

    def __init__(self, xs: np.ndarray, y: np.ndarray) -> None:
        super().__init__()
        assert xs.shape == (y.shape[0], xs.shape[1]), 'length of xs must be equal to length of y'
        self.xs = xs
        self.y = y

    def fitness(self, individual: Individual) -> float:
        dna = individual.genotype.dna
        return -1 * least_squares(self.xs, self.y, dna, fit=polynomial_fit)

    def value(self, individual: Individual, x: np.ndarray) -> float:
        dna = individual.genotype.dna
        return polynomial_fit(dna, x)


class MaximumLikelihood(ObservationBasedFitness):

    def __init__(self, xs: np.ndarray, y: np.ndarray) -> None:
        super().__init__()
        assert xs.shape == (y.shape[0], xs.shape[1]), 'length of xs must be equal to length of y'
        classes = np.unique(y)
        assert len(classes) == 2 and np.array_equal(classes, np.array([0, 1])), 'y-values can only belong [0, 1] discrete interval'
        self.xs = xs
        self.y = y

    def fitness(self, individual: Individual) -> float:
        alpha, *coeff = individual.genotype.dna
        y_positive = np.where(self.y == 1)
        likelihood: float = 0
        for x in self.xs[y_positive]:
            likelihood += logit(alpha, coeff, x)
        y_negative = np.where(self.y == 0)
        for x in self.xs[y_negative]:
            likelihood += 1 - logit(alpha, coeff, x)
        return likelihood

    def value(self, individual: Individual, x: np.ndarray) -> float:
        alpha, *coeff = individual.genotype.dna
        return logit(alpha, coeff, x)



# -*- coding: utf-8 -*-
"""Core Auxein mutations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import ABC, abstractmethod

import numpy as np

from auxein.population import Individual


class Fitness(ABC):

    @abstractmethod
    def fitness(self, individual: Individual) -> float:
        pass

    @abstractmethod
    def value(self, individual: Individual, x: np.ndarray) -> float:
        pass



# -*- coding: utf-8 -*-
"""Core Auxein mutations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import ABC, abstractmethod
from typing import List

import numpy as np


def cumulative_probability_distribution(index: int, probabilities: List[float]) -> float:
    return sum(probabilities[:index + 1])


class Selection(ABC):

    def __init__(self, offspring_size: int) -> None:
        self.__offspring_size = offspring_size
        self.parents_to_select = np.around(np.roots([1, -1, -offspring_size / 2])[0])

    @property
    def offspring_size(self) -> int:
        return self.__offspring_size

    @abstractmethod
    def select(self, individual_ids: List[str], probabilities: List[float]) -> List[str]:
        pass


class StochasticUniversalSampling(Selection):

    def __init__(self, offspring_size: int) -> None:
        super().__init__(offspring_size=offspring_size)

    def select(self, individual_ids: List[str], probabilities: List[float]) -> List[str]:
        index = 0
        mating_pool: List[str] = []
        r = np.random.uniform(0, 1 / self.parents_to_select)
        while len(mating_pool) < self.parents_to_select:
            while r <= cumulative_probability_distribution(index, probabilities):
                mating_pool.append(individual_ids[index])
                r += 1 / self.parents_to_select
            index += 1
        return mating_pool



# -*- coding: utf-8 -*-
"""Core Auxein mutations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from abc import ABC, abstractmethod
from typing import Tuple, List

from auxein.population import Item, Population


class Distribution(ABC):

    @abstractmethod
    def get(self, population: Population) -> List[Tuple[str, float]]:
        pass


class Fps(Distribution):

    def __init__(self) -> None:
        super().__init__()

    def get(self, population: Population) -> List[Tuple[str, float]]:
        return list(map(lambda item: (item.individual.id, item.fitness / population.total_fitness()), population.pool))


class FpsWithWindowing(Distribution):

    def __init__(self) -> None:
        super().__init__()

    def __scale_fitness_function(self, item: Item, minimum_fitness: float) -> float:
        return item.fitness - minimum_fitness

    def get(self, population: Population) -> List[Tuple[str, float]]:
        minimum_fitness = population.min_fitness()
        total_fitness = sum(self.__scale_fitness_function(item, minimum_fitness) for item in population.pool)
        return list(map(lambda item: (item.individual.id, self.__scale_fitness_function(item, minimum_fitness) / total_fitness), population.pool))


class SigmaScaling(Distribution):

    def __init__(self) -> None:
        super().__init__()

    def __scale_fitness_function(self, item: Item, population: Population) -> float:
        max_value: float = max(item.fitness - (population.mean_fitness() - 2 * population.std_fitness()), 0)
        return max_value

    def get(self, population: Population) -> List[Tuple[str, float]]:
        total_fitness = sum(self.__scale_fitness_function(item, population) for item in population.pool)
        return list(map(lambda item: (item[0].id, self.__scale_fitness_function(item, population) / total_fitness), population.pool))
